/*
 * benchmark-sybase.c
 *
 * Interface between mod_benchmark and Sybase (using ct library).
 * (source inspired by sample/ctlibrary/firstapp.c from Sybase)
 *
 * To implement other DB, just create another .so with
 * same functions.
 */
static const char *rcsid = "$Id$";

#include "common.h"
#include <httpd.h>
#include <ctpublic.h>
#include <stdio.h>

typedef struct {
	CS_CONTEXT	*ctxt;
	CS_CONNECTION	*conn;
} opaque_t;

#define CHECK_FAIL(ctxt,ret,str) \
if ((ret) != CS_SUCCEED) { \
	fprintf(stderr,"%s\n",str);\
	if (ctxt != (CS_CONTEXT *)NULL) { \
		ct_exit(ctxt,CS_FORCE_EXIT); \
		cs_ctx_drop(ctxt); \
	} \
	return NULL; \
}

/*
 * Forward declarations:
 */
CS_RETCODE CS_PUBLIC CSMsgCallback(CS_CONTEXT *context, CS_CLIENTMSG *emsgp);
CS_RETCODE CS_PUBLIC SrvMsgCallback(CS_CONTEXT *cp, CS_CONNECTION *chp, CS_SERVERMSG *msgp);
CS_RETCODE CS_PUBLIC CliMsgCallback(CS_CONTEXT *context, CS_CONNECTION *conn, CS_CLIENTMSG *emsgp);
CS_RETCODE Execute(CS_CHAR *query, CS_CONNECTION *conn);

/*
 * Init function: 
 * Allocate some private data and connect to the DB
 */
const void *db_init(apr_pool_t *pool, const char *srv, const char *db, const char *user, const char *pwd)
{
	CS_CONTEXT	*ctxt;
	CS_CONNECTION	*conn;
	CS_RETCODE	ret;
	opaque_t	*opa;
	char		buff[128];

	/* Allocate a context structure: */
	ctxt = (CS_CONTEXT *)NULL;
	ret = cs_ctx_alloc(CS_VERSION_100,&ctxt);
	CHECK_FAIL(ctxt,ret,"cs_ctx_alloc() failed");

	/* Initialize Client Library: */
	ret = ct_init(ctxt,CS_VERSION_100);
	CHECK_FAIL(ctxt,ret,"ct_init() failed");

	/* Install a callback for handling CS-library errors: */
	ret = cs_config(ctxt,CS_SET,CS_MESSAGE_CB,(CS_VOID *)CSMsgCallback,CS_UNUSED,NULL);
	CHECK_FAIL(ctxt,ret,"cs_config() failed");

	/* Install a callback function to handle Client-Library errors: */
	ret = ct_callback(ctxt,NULL,CS_SET,CS_CLIENTMSG_CB,(CS_VOID *)CliMsgCallback);
	CHECK_FAIL(ctxt,ret,"ct_callback() for server failed");

	/* Install a callback function to handle server errors: */
	ret = ct_callback(ctxt,NULL,CS_SET,CS_SERVERMSG_CB,(CS_VOID *)SrvMsgCallback);
	CHECK_FAIL(ctxt,ret,"ct_callback() for server failed");

	/* Allocate a connection structure: */
	ret = ct_con_alloc(ctxt,&conn);
	CHECK_FAIL(ctxt,ret,"ct_con_alloc() failed");

	/* Set the user credentials: */
	ret = ct_con_props(conn,CS_SET,CS_USERNAME,(CS_CHAR *)user,CS_NULLTERM,NULL);
	CHECK_FAIL(ctxt,ret,"ct_con_props() failed for username");

	ret = ct_con_props(conn,CS_SET,CS_PASSWORD,(CS_CHAR *)pwd,CS_NULLTERM,NULL);
	CHECK_FAIL(ctxt,ret,"ct_con_props() failed for password");

	/*
	ret = ct_con_props(conn,CS_SET,CS_HOSTNAME,(CS_CHAR *)srv,CS_NULLTERM,NULL);
	CHECK_FAIL(ctxt,ret,"ct_con_props() failed for hostname");
	*/

	/* Create the connection: */
	ret = ct_connect(conn,(CS_CHAR *)srv,CS_NULLTERM);
	CHECK_FAIL(ctxt,ret,"ct_connect() failed");

	/* Change to the specified DB even if it's the default one for user: */
	snprintf(buff,sizeof(buff)-1,"use %s",db);
	buff[sizeof(buff)-1] = '\0';
	Execute(buff,conn);

	/* Create and return the opaque structure: */
	opa = (opaque_t *)apr_palloc(pool,sizeof(opaque_t));
	opa->ctxt = ctxt;
	opa->conn = conn;

	return opa;
}

/*
 * Called each time a row must be inserted into the Log table:
 */
int db_insert(const void *opaque, const char *query)
{
	CS_RETCODE	ret;

	if ((ret = Execute((CS_CHAR *)query,((opaque_t *)opaque)->conn)) != CS_SUCCEED)
		return 1;

	return 0;
}

/*
 * Called when apache exits:
 */
void db_close(const void *opaque)
{
	CS_RETCODE	ret;

	/* Close the connection and drop its control structure: */
	if ((ret = ct_close(((opaque_t *)opaque)->conn,CS_UNUSED)) != CS_SUCCEED)
		fprintf(stderr,"ct_close() failed\n");

	if ((ret = ct_con_drop(((opaque_t *)opaque)->conn)) != CS_SUCCEED) 
		fprintf(stderr,"ct_con-drop() failed\n");

	/* Tells the Client-Library we are done: */
	if ((ret = ct_exit(((opaque_t *)opaque)->ctxt,CS_UNUSED)) != CS_SUCCEED) 
		fprintf(stderr,"ct_exit() failed\n");

	/* Drop the context structure: */
	if ((ret = cs_ctx_drop(((opaque_t *)opaque)->ctxt)) != CS_SUCCEED)
		fprintf(stderr,"cs_ctx_drop() failed\n");
}

/*
 * Execute a query:
 */
CS_RETCODE Execute(CS_CHAR *query, CS_CONNECTION *conn)
{
	CS_RETCODE	ret;
	CS_COMMAND	*cmd;
	CS_INT		res;

	/* Allocate a command structure: */
	if ((ret = ct_cmd_alloc(conn,&cmd)) != CS_SUCCEED) {
		fprintf(stderr,"ct_cmd_alloc() failed\n");
		return ret;
	}

	/* Build the Query: */
	if ((ret = ct_command(cmd,CS_LANG_CMD,query,CS_NULLTERM,CS_UNUSED)) != CS_SUCCEED) {
		ct_cmd_drop(cmd);
		fprintf(stderr,"ct_command() failed\n");
		return ret;
	}

	/* Send the query: */
	if ((ret = ct_send(cmd)) != CS_SUCCEED) {
		ct_cmd_drop(cmd);
		fprintf(stderr,"ct_send() failed\n");
		return ret;
	}

	while (ct_results(cmd,&res) == CS_SUCCEED);

	/* Free the command structure: */
	if ((ret = ct_cmd_drop(cmd)) != CS_SUCCEED) {
		fprintf(stderr,"ct_cmd_drop() failed\n");
	}

	return ret;
}

/*
** Handler for server messages. Client-Library will call this
** routine when it receives a message from the server.
*/
CS_RETCODE CS_PUBLIC SrvMsgCallback(CS_CONTEXT *cp, CS_CONNECTION *chp, CS_SERVERMSG *msgp)
{
	/*
	** Print the message info.
	*/
	fprintf(stderr,
		"Server message:\n\t");
	fprintf(stderr,
		"number(%ld) severity(%ld) state(%ld) line(%ld)\n",
		(long)msgp->msgnumber, (long)msgp->severity,
		(long)msgp->state, (long)msgp->line);

	/*
	** Print the server name if one was supplied.
	*/
	if (msgp->svrnlen > 0)
		fprintf(stderr, "\tServer name: %s\n", msgp->svrname);

	/*
	** Print the null terminated message.
	*/
	fprintf(stderr, "\t%s\n", msgp->text);

	return CS_SUCCEED;
}

/*
**  Client-Library error handler. This function will be invoked
**  when a Client-Library has detected an error. Before Client-
**  Library routines return CS_FAIL, this handler will be called
**  with additional error information.
*/
CS_RETCODE CS_PUBLIC CliMsgCallback(CS_CONTEXT *context, CS_CONNECTION *conn, CS_CLIENTMSG *emsgp)
{

	/*
	** Error number: Print the error's severity, number, origin, and
	** layer. These four numbers uniquely identify the error.
	*/
	fprintf(stderr,
		"Client Library error:\n\t");
	fprintf(stderr,
		"severity(%ld) number(%ld) origin(%ld) layer(%ld)\n",
		(long)CS_SEVERITY(emsgp->severity),
		(long)CS_NUMBER(emsgp->msgnumber),
		(long)CS_ORIGIN(emsgp->msgnumber),
		(long)CS_LAYER(emsgp->msgnumber));

	/*
	** Error text: Print the error text.
	*/
	fprintf(stderr, "\t%s\n", emsgp->msgstring);

	/*
	** Operating system error information: Some errors, such as
	** network errors, may have an operating system error associated
	** with them. If there was an operating system error, this code
	** prints the error message text.
	*/
	if (emsgp->osstringlen > 0)
	{
		fprintf(stderr,
			"Operating system error number(%ld):\n",
			(long)emsgp->osnumber);
		fprintf(stderr, "\t%s\n", emsgp->osstring);
	}

	/*
	** If we return CS_FAIL, Client-Library marks the connection as
	** dead. This means that it cannot be used anymore. If we return
	** CS_SUCCEED, the connection remains alive if it was not already
	** dead.
	*/
	return CS_SUCCEED;
}

/*
**  CS-Library error handler. This function will be invoked
**  when CS-Library has detected an error.
*/
CS_RETCODE CS_PUBLIC CSMsgCallback(CS_CONTEXT *context, CS_CLIENTMSG *emsgp)
{
	/*
	** Print the error number and message.
	*/
	fprintf(stderr,
		"CS-Library error:\n");
	fprintf(stderr,
		"\tseverity(%ld) layer(%ld) origin(%ld) number(%ld)",
		(long)CS_SEVERITY(emsgp->msgnumber),
		(long)CS_LAYER(emsgp->msgnumber),
		(long)CS_ORIGIN(emsgp->msgnumber),
		(long)CS_NUMBER(emsgp->msgnumber));

	fprintf(stderr, "\t%s\n", emsgp->msgstring);

	/*
	** Print any operating system error information.
	*/
	if (emsgp->osstringlen > 0)
	{
		fprintf(stderr, "Operating System Error: %s\n",
			emsgp->osstring);
	}

	return CS_SUCCEED;
}

/*
 * $Log:$
 */

/* EOF */
