/*
 * ==========================================================================================
 * mod_benchmark.c
 *
 * Date: 20 Sept 2002
 * Author: Jerome Delamarche
 * Version: 1.5 
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 *	20 Sept 2002:
 *		handling of "ben_srvport" field
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO:
 *
 * ------------------------------------------------------------------------------------------
 *
 * Example usage:
 *
 * LoadModule benchmark_module /apache/dir/mod_benchmark.so
 * AddModule mod_benchmark.c
 *
 * LogFormat "%t %{BenchmarkElapsed}n %{BenchmarkScoreBoard}n \"%r\" %s %B" benchmark_log
 * CustomLog logs/benchmark.log benchmark_log
 * BenchmarkTrace debug
 * BenchmarkUseScoreBoard on|off
 * BenchmarkStore path_to_db_interface.so:host:db:user:pwd
 * BenchmarkWithSysInfo on|off:statprogram:socketname:interval
 * BenchmarkAlertAction proc	// will receive the URL and the elapsed time as parameters 
 *
 * <Location>
 *	BenchmarkEngine on|off
 *	BenchmarkAccept \.html$|\.php$
 * 	BenchmarkAlertLimit milliseconds
 * </Location>
 *
 * ==========================================================================================
 */
#include "common.h"

#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "scoreboard.h"

#include <sys/time.h>
#include <sys/un.h>
#include <signal.h>
#include "sysstat2.h"

static const char *rcsid = "$Id$";

/* private configuration: 
 */
typedef struct {
	int	isOn;		/* directive BenchmarkEngine on|off */
	int	useScoreBoard;	/* directive BenchmarkUseScoreBoard on|off */
	int	withSysInfo;	/* directive BenchmarkWithSysInfo on|off */
	pid_t	statpgmpid;
	char	*socketname;
	int	interval;
	int	traceDebug;	/* directive BenchmarkTrace debug */
	regex_t **regAccept;	/* from BenchmarkAccept directive */
	int 	regCount;	/* count od regex directives */
	apr_dso_handle_t *hmod;		/* handle to the DB module */
	void*   (*db_init)(apr_pool_t *pool, char *srv, char *db, char *u, char *p);	/* DB module db_init() entry point */
	int	(*db_insert)(void *opaque, const char *query);				/* DB module db_insert() entry point */
	void   	(*db_close)(void *opaque);						/* DB module db_close() entry point */
	void    *db_opaque;	/* opaque descriptor returned by db_init() */
	char	*alertAction;	/* directive BenchmarkAlertAction */
	int	alertLimit;	/* directive BenchmarkAlertLimit */
} bench_config;

/* Module private variables */
static int gServerLimit;
static int gThreadLimit;

/* Forward declaration... 
 */
module benchmark_module;

/* 
 * ------------------
 * Directive handlers: 
 * ------------------
 */
#if WITH_APACHE2
static int benchmark_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
#endif
#if WITH_APACHE1
static void benchmark_init(server_rec *s, apr_pool_t *p)
#endif
{
	if (!ap_exists_scoreboard_image()) {
		ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ERR,ALE2(APR_SUCCESS)NULL,
			"Server status unavailable in inetd mode");
	}

#if WITH_APACHE2
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS,&gServerLimit);
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS,&gThreadLimit);

	return APR_SUCCESS;
#endif
}

/*
 * -----------------------------
 * Directives at Directory Level:
 * -----------------------------
 */
static const char *benchmark_engine_cmd(cmd_parms *parms, void *config, int bool)
{
	bench_config *cfg = (bench_config *)config;
	cfg->isOn = bool;

	return NULL;
}

static const char *benchmark_accept_cmd(cmd_parms *parms, void *config, char *expr)
{	
	char *msg = NULL;
	bench_config *cfg = (bench_config *)config;
	regex_t *reg,**newreg;

	if (!(reg = ap_pregcomp(parms->pool,expr,REG_EXTENDED))) 
		return "Bad RE for BenchmarkAccept directive";

	/* allocate more room : */
	newreg = (regex_t **)apr_palloc(parms->pool,sizeof(regex_t *) * (cfg->regCount + 1));
	if (cfg->regCount) 
		memcpy(newreg,cfg->regAccept,sizeof(regex_t *) * cfg->regCount);
	newreg[cfg->regCount] = reg;
	cfg->regAccept = newreg;
	cfg->regCount++;

	return NULL;
}

static const char *benchmark_alertlimit_cmd(cmd_parms *parms, void *config, char *proc)
{
	bench_config *cfg = (bench_config *)config;

	cfg->alertLimit = atoi(proc);
	if (cfg->alertLimit <= 0)
		return "Value for BenchmarkAlertLimit must be > 0";

	return NULL;
}

/*
 * --------------------------
 * Directives at Server Level:
 * --------------------------
 */
static const char *benchmark_trace_cmd(cmd_parms *parms, void *config, char *cmd)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);

	if (strstr(cmd,"debug"))
		cfg->traceDebug = 1;

	return NULL;
}

static const char *benchmark_scoreboard_cmd(cmd_parms *parms, void *dummy, int bool)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);
	cfg->useScoreBoard = bool;

	return NULL;
}

/*
 * Cleanup routine when statProgram is launched by benchmark_sysinfo_cmd():
 */
#if WITH_APACHE2
static apr_status_t benchmark_childproc_cleanup(void *v)
#else
static void benchmark_childproc_cleanup(void *v)
#endif
{
	bench_config *cfg = (bench_config *)v;

	ap_log_error(APLOG_MARK,APLOG_INFO|APLOG_NOERRNO,APR_SUCCESS,NULL,"childproc_cleanup");

	if (cfg->statpgmpid != -1)
		kill(cfg->statpgmpid,SIGINT);

#if WITH_APACHE2
	return APR_SUCCESS;
#endif
}

static const char *benchmark_sysinfo_cmd(cmd_parms *parms, void *dummy, char *sysinfo)
{
	/* Split sysinfo into 4 fields: on/off, statpgm, socketname, interval */
	char *onoff,*statpgm,*socketname,*interval;
	char *msg = "Bad syntax. Some fields are missing";

	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);

	onoff = sysinfo;
	if (!(statpgm = strchr(onoff,':'))) return msg;
	*statpgm = 0;
	statpgm++;

	if (!(socketname = strchr(statpgm,':'))) return msg;
	*socketname = 0;
	socketname++;

	if (!(interval = strchr(socketname,':'))) return msg;
	*interval = 0; 
	interval++;

	/* check values :*/
	if (!strcasecmp(onoff,"on") || !strcmp(onoff,"1") || !strcasecmp(onoff,"true")) {
		if (atoi(interval) < 1) 
			return "Bad value for interval";

		cfg->withSysInfo = 1;
		cfg->statpgmpid  = -1;
		cfg->socketname  = socketname;
		cfg->interval    = atoi(interval);

		{
			/* try to start the Stat program: */
			cfg->statpgmpid = fork();
			if (cfg->statpgmpid < 0) 
				return "Could not fork the Stat program";

			if (cfg->statpgmpid == 0) {
				/* child executes the stat pgm: */
				if (execl(statpgm,statpgm,socketname,interval,NULL) < 0) 
					ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
						"Could not exec() the Stat program");
				exit(1);
			}

			/* father register a cleanup routine to kill the child: */
			apr_pool_cleanup_register(parms->pool,cfg,benchmark_childproc_cleanup,apr_pool_cleanup_null);
		}
	}

	return NULL;
}

static const char *benchmark_alertaction_cmd(cmd_parms *parms, void *dummy, char *proc)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);
	cfg->alertAction = proc;

	return NULL;
}

/*
 * Cleanup routine when Database storing is required by benchmark_store_cmd():
 */
#if WITH_APACHE2
static apr_status_t benchmark_dbclose_cleanup(void *v)
#else
static void benchmark_dbclose_cleanup(void *v)
#endif
{
	bench_config *cfg = (bench_config *)v;

	ap_log_error(APLOG_MARK,APLOG_INFO|APLOG_NOERRNO,ALE2(APR_SUCCESS)NULL,"dbclose_cleanup");

	cfg->db_close(cfg->db_opaque);

#if WITH_APACHE2
	/* +++++ added on the 12 Sep 2002 */
	apr_dso_unload(cfg->hmod);

	return APR_SUCCESS;
#endif
}

static const char *benchmark_store_cmd(cmd_parms *parms, void *config, char *storedesc)
{
	/* Split storedesc into 5 fields: dbtype,server,dbname,user,pwd: */
	char *server,*dbname, *user, *pwd;
	char *msg = "Bad syntax. Some fields are missing";
	char modname[128];
	char *fullmodname;
	apr_status_t apr;

	bench_config *cfg = ap_get_module_config(parms->server->module_config,&benchmark_module);

	if (!(server = strchr(storedesc,':'))) return msg;
	*server = 0;
	server++;

	if (!(dbname = strchr(server,':'))) return msg;
	*dbname = 0; 
	dbname++;

	if (!(user = strchr(dbname,':'))) return msg;
	*user = 0;
	user++;

	if (!(pwd = strchr(user,':'))) return msg;
	*pwd = 0;
	pwd++;

	/* Load the DB interface: */
#if WITH_APACHE2
	if ((apr = apr_dso_load(&cfg->hmod,storedesc,parms->pool)) != APR_SUCCESS) {
		char buff[512];
		apr_dso_error(cfg->hmod,buff,sizeof(buff));
		ap_log_error(APLOG_MARK,APLOG_ERR,apr,parms->server,buff);
		ap_log_error(APLOG_MARK,APLOG_ERR,apr,parms->server,
			"Module %s not found",storedesc);
		return "DB module not found (see error log)";
	}

	/* Get the entry-point: */
	if ((apr = apr_dso_sym((apr_dso_handle_sym_t *)&cfg->db_init,cfg->hmod,"db_init")) != APR_SUCCESS) {
		char buff[512];
		apr_dso_error(cfg->hmod,buff,sizeof(buff));
		ap_log_error(APLOG_MARK,APLOG_ERR,apr,parms->server,buff);
		return "db_init not found in DB module";
	}
	if ((apr = apr_dso_sym((apr_dso_handle_sym_t *)&cfg->db_insert,cfg->hmod,"db_insert")) != APR_SUCCESS) 
		return "db_insert not found in DB module";
	if ((apr = apr_dso_sym((apr_dso_handle_sym_t *)&cfg->db_close,cfg->hmod,"db_close")) != APR_SUCCESS) 
		return "db_close not found in DB module";
#endif

#if WITH_APACHE1
	/* Load the DB interface: */
	if (!(cfg->hmod = ap_os_dso_load(storedesc))) {
		ap_log_error(APLOG_MARK,APLOG_ERR,parms->server,
			"Module %s not found",storedesc);
		return "DB module not found (see error log)";
	}

	/* Get the entry-point: */
	if (!(cfg->db_init = dlsym(cfg->hmod,"db_init"))) 
		return "db_init not found in DB module";
	if (!(cfg->db_insert = dlsym(cfg->hmod,"db_insert")))
		return "db_insert not found in DB module";
	if (!(cfg->db_close = dlsym(cfg->hmod,"db_close")))
		return "db_close not found in DB module";
#endif

	/* Call db_init() ! */
	if (!(cfg->db_opaque = cfg->db_init(parms->pool,server,dbname,user,pwd)))
		return "Database initialization failed";

	/* Everything OK, register the db_close() call: */
	apr_pool_cleanup_register(parms->pool,cfg,benchmark_dbclose_cleanup,apr_pool_cleanup_null);

	return NULL;
}

/* 
 * ------------------------
 * Module Apache directives:
 * ------------------------
 */
command_rec benchmark_cmds[] = {
    { "BenchmarkEngine", benchmark_engine_cmd, NULL, ACCESS_CONF, FLAG, "Start/Stop the engine" },
    { "BenchmarkUseScoreBoard", benchmark_scoreboard_cmd, NULL, RSRC_CONF, FLAG, "Use APACHE ScoreBoard ?" },
    { "BenchmarkWithSysInfo", benchmark_sysinfo_cmd, NULL, RSRC_CONF, TAKE1, "Store the system average load and mem usage ?" },
    { "BenchmarkTrace", benchmark_trace_cmd, NULL, RSRC_CONF, TAKE1, "Debug listed components (debug)" },
    { "BenchmarkAccept", benchmark_accept_cmd, NULL, ACCESS_CONF, TAKE1, "Extended RE that list the files to be analyzed" },
    { "BenchmarkStore", benchmark_store_cmd, NULL, RSRC_CONF, TAKE1, "Optional definition of storage target (dbtype:dbname:user:pwd)" },
    { "BenchmarkAlertAction", benchmark_alertaction_cmd, NULL, RSRC_CONF, TAKE1, "Process to call for sending alerts" },
    { "BenchmarkAlertLimit", benchmark_alertlimit_cmd, NULL, ACCESS_CONF, TAKE1, "Limit of time (in milliseconds) a request can take before alert" },
    { NULL }
};

/*
 * --------------------------------
 * Directory Configuration Callback:
 * --------------------------------
 */
static void *create_dir_config(apr_pool_t *p, char *path)
{
	bench_config *cfg = (bench_config *)apr_pcalloc(p,sizeof(bench_config));

	memset(cfg,0,sizeof(bench_config));

	return (void *)cfg;
}

/*
 * --------------------------------
 * Directory Configuration Callback:
 * --------------------------------
 */
static void *create_server_config(apr_pool_t *p, server_rec *r)
{
	bench_config *cfg = (bench_config *)apr_pcalloc(p,sizeof(bench_config));

	memset(cfg,0,sizeof(bench_config));

	return (void *)cfg;
}

/*
 * ------------------------------
 * Start benchmarking the request:
 * ------------------------------
 */
static int benchmark_start(request_rec *r)
{
	struct timeval stamp;
	char buff[64];
	int idx;
	bench_config *scf = (bench_config *)ap_get_module_config(r->server->module_config,&benchmark_module);
	bench_config *cfg = (bench_config *)ap_get_module_config(r->per_dir_config,&benchmark_module);

	if (scf->traceDebug)
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"benchmark_start on URI: %s",r->uri);

	/* benchmarking on ? */
	if (!cfg->isOn) 
		return DECLINED;

	if (scf->traceDebug) 
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"Benchmark Engine on for URI: %s",r->uri);

	/* matches the regex ? */
	for ( idx = 0 ; idx < cfg->regCount ; idx++ ) {
		if (regexec(cfg->regAccept[idx],r->uri,0,NULL,0))
			continue;

		/* we pass info to BenchmarkStop using headers_in because of the subquery problem */
		gettimeofday(&stamp, NULL);
		sprintf(buff,"%ld.%ld",stamp.tv_sec,stamp.tv_usec);
		apr_table_set(r->headers_in, "BenchmarkStart", apr_pstrdup(r->pool,buff));

		if (scf->traceDebug) 
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"Benchmark started for %s",r->uri);

		return DECLINED;
	}

    	return DECLINED;
}

/*
 * ------------------------
 * Log the benchmark result:
 * ------------------------
 */
static int benchmark_stop(request_rec *r)
{
	long took_sec, took_usec;
	struct timeval stamp_begin,stamp_end;
	char buff[64];
	const char *stamp_buff;
	int server_ready = 0;
	int server_busy = 0;
	sysstat2_t as;
	bench_config *scf = (bench_config *)ap_get_module_config(r->server->module_config,&benchmark_module);
	bench_config *cfg = (bench_config *)ap_get_module_config(r->per_dir_config,&benchmark_module);
	char buftime[32];
	time_t curtime = time(NULL);
	strftime(buftime,sizeof(buftime),"%Y-%m-%d %H:%M:%S",localtime(&curtime));

	/* +++++ removed on the 16 Sep 2002:
	 * +++++ when rewriting occurs, start() could have been done and the
	 * +++++ translated URI, but for the original one, "isOn" is FALSE !
	if (!cfg->isOn)
		return DECLINED;
	*/

	gettimeofday(&stamp_end, NULL);

	/* benchmark_start called ? (regex...) */
	if (!(stamp_buff = apr_table_get(r->headers_in, "BenchmarkStart")))
		return DECLINED;

	if (scf->traceDebug)
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"BenchmarkStop");

	sscanf(stamp_buff,"%ld.%ld",&stamp_begin.tv_sec,&stamp_begin.tv_usec);

	took_sec = stamp_end.tv_sec - stamp_begin.tv_sec;
	took_usec = stamp_end.tv_usec - stamp_begin.tv_usec;
	if (took_usec < 0) {
	    --took_sec;
	    took_usec += 1000000;
	}

	/* for logging purpose... */
	sprintf(buff, "%ld.%ld", took_sec, took_usec);
	apr_table_set(r->notes,"BenchmarkElapsed",buff);

	/* use scoreboard ? */
	if (scf->useScoreBoard) {
		int i,j,status;
#if WITH_APACHE2
		process_score *ps_record;
#else
		short_score score_record;
#endif
		char buff[64];

		if (scf->traceDebug)
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"Score board used");

#if WITH_APACHE2
		for ( i = 0 ; i < gServerLimit ; i++ ) {
			ps_record = ap_get_scoreboard_process(i);
			for ( j = 0 ; j < gThreadLimit ; j++ ) {
				worker_score *ws_record = ap_get_scoreboard_worker(i,j);
				status = ws_record->status;

				if (!ps_record->quiescing && ps_record->pid) {
					if (status == SERVER_READY && ps_record->generation == ap_my_generation)
						server_ready++;
					else if (status != SERVER_DEAD &&
						 status != SERVER_STARTING &&
						 status != SERVER_IDLE_KILL)
						server_busy++;
				}
			}
		}
#endif
#if WITH_APACHE1
		ap_sync_scoreboard_image();
		for ( i = 0 ; i < HARD_SERVER_LIMIT ; i++ ) {
			score_record = ap_scoreboard_image->servers[i];
			status = score_record.status;
			if (status == SERVER_READY)
				server_ready++;
			else if (status != SERVER_DEAD)
				server_busy++;
		}
#endif

		sprintf(buff, "%ld.%ld", server_ready, server_busy);
		apr_table_set(r->notes,"BenchmarkScoreBoard",buff);
	}

	/* use sysinfo ? */
	memset(&as,0,sizeof(as));
	if (scf->withSysInfo) {
		int s;

		if (scf->traceDebug)
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"with sys info");

		/* connect to the socket to get the system stats: */
		if ((s = socket(AF_LOCAL,SOCK_STREAM,0)) < 0) {
			ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)r->server,"socket() failed");
		}
		else {
			struct sockaddr_un addr;
			memset(&addr,0,sizeof(addr));
			addr.sun_family = AF_LOCAL;
			strncpy(addr.sun_path,scf->socketname,sizeof(addr.sun_path) - 1);
			addr.sun_path[sizeof(addr.sun_path)-1] = '\0';

			if (connect(s,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
				ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)r->server,"connect() failed");
			}
			else {
				char *req = "?";
				send(s,req,1,0);
				if (recv(s,&as,sizeof(as),0) != sizeof(as)) {
					ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)r->server,"recv() failed");
				}
			}

			close(s);
		}
	}

	/* Log in Database ? */
	scf = (bench_config *)ap_get_module_config(r->server->module_config,&benchmark_module);
	if (scf->hmod) {
		const char *query;

		unsigned short port;
#if WITH_APACHE1
		port = ntohs(r->connection->local_addr.sin_port);
#endif
#if WITH_APACHE2
		port = ntohs(r->connection->local_addr->port);
#endif

		query = apr_psprintf(r->pool,
				"INSERT INTO t_benchmark_log "
				"(log_date,url_value,arg_value,fil_value,log_status,ben_sec,ben_usec,ben_ready,ben_busy,"
				" ben_pzombie,ben_pstopped,ben_prunning,ben_psleeping,ben_avgload,"
				" ben_memuse,ben_swapuse,ben_cliaddr,ben_srvaddr,ben_srvport) "
				"VALUES ('%s','%s','%s','%s',%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)",
				buftime,(r->uri)?r->uri:"",(r->args)?r->args:"",(r->filename)?r->filename:"",
				r->status,took_sec,took_usec,server_ready,server_busy,
				as.zombie,as.stopped,as.running,as.sleeping,as.avgload,as.memuse,as.swapuse,
				ntohl(inet_addr(r->connection->remote_ip)),ntohl(inet_addr(r->connection->local_ip)),port);

		scf->db_insert(scf->db_opaque,query);
	}

	/* Alert action specified ? */
	if (scf->alertAction && cfg->alertLimit > 0 && (took_sec * 1000 + took_usec / 1000) > cfg->alertLimit) {
		char *cmd = apr_psprintf(r->pool,"%s %s %d",scf->alertAction,r->uri,(took_sec * 1000 + took_usec / 1000));
		system(cmd);

		if (scf->hmod) {
			const char *query = apr_psprintf(r->pool,
					"INSERT INTO t_benchmark_alert "
					"(log_date,url_value,arg_value,fil_value,ben_sec,ben_usec) "
					"VALUES ('%s','%s','%s','%s',%d,%d,%d,%d)",
					buftime,(r->uri)?r->uri:"",(r->args)?r->args:"",(r->filename)?r->filename:"",took_sec,took_usec,
					ntohl(inet_addr(r->connection->remote_ip)),ntohl(inet_addr(r->connection->local_ip)));
			scf->db_insert(scf->db_opaque,query);
		}
	}

	return DECLINED;
}

#if WITH_APACHE2
/*
 * ------------------
 * Hooks Registration:
 * ------------------
 */
static void register_hooks(apr_pool_t *pool)
{
	ap_log_error(APLOG_MARK,APLOG_ERR,APR_SUCCESS,NULL,"register mod_benchmark hooks");
	ap_hook_post_config(benchmark_init,NULL,NULL,APR_HOOK_LAST);
	/*
	 * Modified by JD on the 12 Feb 2003:
	 * APR_HOOK_FIRST instead of LAST ! (fix a bug in Apache 2.0.44+)
	 */
	ap_hook_translate_name(benchmark_start,NULL,NULL,APR_HOOK_FIRST);
	ap_hook_log_transaction(benchmark_stop,NULL,NULL,APR_HOOK_FIRST);
}

/*
 * -----------------
 * Module Definition:
 * -----------------
 */
module AP_MODULE_DECLARE_DATA benchmark_module = {
    STANDARD20_MODULE_STUFF,
    create_dir_config,			/* dir config creater */
    NULL,				/* dir merger default is to override */
    create_server_config,		/* server config */
    NULL,				/* merge server config */
    benchmark_cmds,			/* command table */
    register_hooks
};
#endif

#if WITH_APACHE1
module MODULE_VAR_EXPORT benchmark_module = {
    STANDARD_MODULE_STUFF,
    benchmark_init,			/* initializer */
    create_dir_config,			/* dir config creater */
    NULL,				/* dir merger default is to override */
    create_server_config,		/* server config */
    NULL,				/* merge server config */
    benchmark_cmds,			/* command table */
    NULL,				/* handlers */
    benchmark_start,			/* filename translation */
    NULL,				/* check_user_id */
    NULL,				/* check auth */
    NULL,				/* check access */
    NULL,				/* type_checker */
    NULL,				/* fixups */
    benchmark_stop,			/* logger */
    NULL,				/* header parser */
    NULL,				/* child_init */
    NULL,				/* child_exit */
    NULL				/* post read-request */
};
#endif

/*
 * $Log:$
 */

/* EOF */
